# pylint: disable=missing-docstring
import io
import tempfile
from typing import Mapping
from unittest import mock

from django.test.client import Client
from gbpcli import GBP
from requests import PreparedRequest, Response
from requests.adapters import BaseAdapter
from requests.structures import CaseInsensitiveDict
from unittest_fixtures import FixtureContext, FixtureOptions, Fixtures, depends

from gbp_ps.repository import Repo, RepositoryType, sqlite
from gbp_ps.settings import Settings


def tempdir(_options: FixtureOptions, _fixtures: Fixtures) -> FixtureContext[str]:
    with tempfile.TemporaryDirectory() as tempdir_:
        yield tempdir_


@depends("tempdir")
def environ(
    options: FixtureOptions, fixtures: Fixtures
) -> FixtureContext[dict[str, str]]:
    new_environ = options.get("environ", {}).copy()
    new_environ["GBP_PS_SQLITE_DATABASE"] = f"{fixtures.tempdir}/db.sqlite"
    with mock.patch.dict("os.environ", new_environ, clear=True):
        yield new_environ


@depends("environ")
def settings(_options: FixtureOptions, _fixtures: Fixtures) -> Settings:
    return Settings.from_environ()


@depends("settings")
def repo(_options: FixtureOptions, fixtures: Fixtures) -> RepositoryType:
    return Repo(fixtures.settings)


def gbp(options: FixtureOptions, _fixtures: Fixtures) -> GBP:
    url = options.get("gbp", {}).get("url", "http://gbp.invalid/")
    gbp_ = GBP(url)
    gbp_.query._session.mount(  # pylint: disable=protected-access
        url, DjangoToRequestsAdapter()
    )

    return gbp_


@depends("tempdir")
def tempdb(_options: FixtureOptions, fixtures: Fixtures) -> str:
    return f"{fixtures.tempdir}/processes.db"


@depends(tempdb)
def repo_fixture(
    _options: FixtureOptions, fixtures: Fixtures
) -> sqlite.SqliteRepository:
    return sqlite.SqliteRepository(Settings(SQLITE_DATABASE=fixtures.tempdb))


class DjangoToRequestsAdapter(BaseAdapter):  # pylint: disable=abstract-method
    """Requests Adapter to call Django views"""

    def send(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        request: PreparedRequest,
        stream: bool = False,
        timeout: None | float | tuple[float, float] | tuple[float, None] = None,
        verify: bool | str = True,
        cert: None | bytes | str | tuple[bytes | str, bytes | str] = None,
        proxies: Mapping[str, str] | None = None,
    ) -> Response:
        assert request.method is not None
        django_response = Client().generic(
            request.method,
            request.path_url,
            data=request.body,
            content_type=request.headers["Content-Type"],
            **request.headers,
        )

        requests_response = Response()
        requests_response.raw = io.BytesIO(django_response.content)
        requests_response.raw.seek(0)
        requests_response.status_code = django_response.status_code
        requests_response.headers = CaseInsensitiveDict(django_response.headers)
        requests_response.encoding = django_response.get("Content-Type", None)
        requests_response.url = str(request.url)
        requests_response.request = request

        return requests_response



"""gbp-ps tests"""

# pylint: disable=missing-docstring
import datetime as dt
import io
from functools import wraps
from typing import Any, Callable, Iterable, TypeVar

import rich.console
from django.test import TestCase as DjangoTestCase
from gbpcli.theme import DEFAULT_THEME
from gbpcli.types import Console
from rich.theme import Theme
from unittest_fixtures import BaseTestCase

from gbp_ps.repository import Repo, add_or_update_process
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess

LOCAL_TIMEZONE = dt.timezone(dt.timedelta(days=-1, seconds=61200), "PDT")


class TestCase(DjangoTestCase, BaseTestCase):
    """Custom TestCase for gbp-ps tests"""


def make_build_process(**kwargs: Any) -> BuildProcess:
    """Create (and save) a BuildProcess"""
    settings = Settings.from_environ()
    add_to_repo = kwargs.pop("add_to_repo", True)
    update_repo = kwargs.pop("update_repo", False)
    attrs: dict[str, Any] = {
        "build_host": "jenkins",
        "build_id": "1031",
        "machine": "babette",
        "package": "sys-apps/systemd-254.5-r1",
        "phase": "compile",
        "start_time": dt.datetime(2023, 11, 11, 12, 20, 52, tzinfo=dt.timezone.utc),
    }
    attrs.update(**kwargs)
    build_process = BuildProcess(**attrs)

    if add_to_repo:
        repo = Repo(settings)
        if update_repo:
            add_or_update_process(repo, build_process)
        else:
            repo.add_process(build_process)

    return build_process


def string_console() -> tuple[Console, io.StringIO, io.StringIO]:
    """StringIO Console"""
    out = io.StringIO()
    err = io.StringIO()

    return (
        Console(
            out=rich.console.Console(file=out, width=88, theme=Theme(DEFAULT_THEME)),
            err=rich.console.Console(file=err),
        ),
        out,
        err,
    )


T = TypeVar("T")


def parametrized(lists_of_args: Iterable[Iterable[Any]]) -> Callable[..., Any]:
    """Parameterized test"""

    def dec(func: Callable[..., Any]) -> Callable[[TestCase, T], None]:
        @wraps(func)
        def wrapper(self: TestCase, *args: Any, **kwargs: Any) -> None:
            for list_of_args in lists_of_args:
                name = ",".join(str(i) for i in list_of_args)
                with self.subTest(name):
                    func(self, *args, *list_of_args, **kwargs)

        return wrapper

    return dec



"""Helper utilities"""

import datetime as dt
import os
from dataclasses import dataclass, fields
from pathlib import Path
from typing import Any, ClassVar, Self, Sequence

from gbpcli.render import LOCAL_TIMEZONE

FALSE_VALUES = {"0", "f", "false", "n", "no", "off"}
TRUE_VALUES = {"1", "on", "t", "true", "y", "yes"}


# XXX: copied from gentoo-build-publisher  # pylint: disable=fixme
@dataclass(frozen=True)
class BaseSettings:
    """Base class for Settings"""

    # Subclasses should define me as the prefix for environment variables for these
    # settings. For example if prefix is "BUILD_PUBLISHER_" and the field is named "FOO"
    # then the environment variable for that field is "BUILD_PUBLISHER_FOO"
    env_prefix: ClassVar = ""

    @classmethod
    def from_dict(cls: type[Self], prefix: str, data_dict: dict[str, Any]) -> Self:
        """Return Settings instantiated from a dict"""
        params: dict[str, Any] = {}
        value: Any
        for field in fields(cls):
            if (key := f"{prefix}{field.name}") not in data_dict:
                continue

            match field.type:
                case "bool":
                    value = get_bool(data_dict[key])
                case "int":
                    value = int(data_dict[key])
                case "Path":
                    value = Path(data_dict[key])
                case _:
                    value = data_dict[key]

            params[field.name] = value
        return cls(**params)

    @classmethod
    def from_environ(cls: type[Self], prefix: str | None = None) -> Self:
        """Return settings instantiated from environment variables"""
        if prefix is None:
            prefix = cls.env_prefix

        return cls.from_dict(prefix, dict(os.environ))


def get_today() -> dt.date:
    """Return today's date"""
    return dt.datetime.now().astimezone(LOCAL_TIMEZONE).date()


def format_timestamp(timestamp: dt.datetime) -> str:
    """Format the timestamp as a string

    Like render.from_timestamp(), but if the date is today's date then only display the
    time. If the date is not today's date then only return the date.
    """
    if (date := timestamp.date()) == get_today():
        return f"[timestamp]{timestamp.strftime('%X')}[/timestamp]"
    return f"[timestamp]{date.strftime('%b%d')}[/timestamp]"


def get_bool(value: str | bytes | bool) -> bool:
    """Return the boolean value of the truthy/falsey string"""
    if isinstance(value, bool):
        return value

    if isinstance(value, bytes):
        value = value.decode("UTF-8")

    if value.lower() in FALSE_VALUES:
        return False

    if value.lower() in TRUE_VALUES:
        return True

    raise ValueError(value)


def find(item: Any, items: Sequence[Any]) -> int:
    """Return the index of the first item in items

    If item is not found in items, return -1.
    """
    try:
        return items.index(item)
    except ValueError:
        return -1
